from __future__ import annotations
from typing import List, Optional, TYPE_CHECKING

from .base import Interaction

from ..components import ActionRow, Component
from ..utils import cached_slot_property
from ..message import Message

__all__ = (
    'MessageInteraction',
    'MessageInteractionData'
)

if TYPE_CHECKING:
    from ..types.interactions import (
        Interaction as InteractionPayload,
        ComponentInteractionData as ComponentInteractionDataPayload,
        ComponentType
    )
    from ..state import ConnectionState


class MessageInteraction(Interaction):
    """Represents an interaction with a message component.

    Current examples are buttons and dropdowns.

    .. versionadded:: 2.1

    Attributes
    -----------
    id: :class:`int`
        The interaction's ID.
    type: :class:`InteractionType`
        The interaction type.
    guild_id: Optional[:class:`int`]
        The guild ID the interaction was sent from.
    channel_id: Optional[:class:`int`]
        The channel ID the interaction was sent from.
    application_id: :class:`int`
        The application ID that the interaction was for.
    author: Optional[Union[:class:`User`, :class:`Member`]]
        The user or member that sent the interaction.
    message: Optional[:class:`Message`]
        The message that sent this interaction.
    component: :class:`Component`
        The component the user interacted with
    values: Optional[List[:class:`str`]]
        The values the user selected
    token: :class:`str`
        The token to continue the interaction. These are valid
        for 15 minutes.
    data: :class:`MessageInteractionData`
        The wrapped interaction data.
    """

    def __init__(self, *, data: InteractionPayload, state: ConnectionState):
        super().__init__(data=data, state=state)
        self.data = MessageInteractionData(data=data.get('data'))
        self.message = Message(state=self._state, channel=self.channel, data=data['message'])
    
    @property
    def values(self) -> Optional[List[str]]:
        return self.data.values

    @cached_slot_property('_cs_component')
    def component(self) -> Component:
        for action_row in self.message.components:
            if not isinstance(action_row, ActionRow):
                continue
            for component in action_row.children:
                if component.custom_id == self.data.custom_id:
                    return component


class MessageInteractionData:
    """Represents the data of an interaction with a message component.

    .. versionadded:: 2.1

    Attributes
    ----------
    id: :class:`int`
        The application command ID.
    name: :class:`str`
        The application command name.
    type: :class:`ApplicationCommandType`
        The application command type.
    custom_id: :class:`str`
        The custom ID of the component.
    component_type: :class:`ComponentType`
        The type of the component.
    values: Optional[List[:class:`str`]]
        The values the user selected
    """

    __slots__ = (
        'custom_id',
        'component_type',
        'values'
    )

    def __init__(self, *, data: ComponentInteractionDataPayload):
        data = {} if data is None else data
        self.custom_id: str = data.get('custom_id')
        self.component_type: ComponentType = data.get('component_type')
        self.values: Optional[List[str]] = data.get('values')



from __future__ import annotations
from typing import Any, Dict, Optional, TYPE_CHECKING, Union

from .base import Interaction

from ..channel import _threaded_channel_factory
from ..enums import OptionType, ApplicationCommandType, try_enum
from ..guild import Guild
from ..role import Role
from ..user import User
from ..member import Member
from ..message import Message

__all__ = (
    'ApplicationCommandInteraction',
    'ApplicationCommandInteractionData',
    'ApplicationCommandInteractionDataResolved'
)

if TYPE_CHECKING:
    from ..types.interactions import (
        Interaction as InteractionPayload,
        ApplicationCommandInteractionData as ApplicationCommandInteractionDataPayload,
        ApplicationCommandInteractionDataResolved as ApplicationCommandInteractionDataResolvedPayload
    )
    from ..state import ConnectionState
    from ..channel import VoiceChannel, StageChannel, TextChannel, CategoryChannel, StoreChannel, PartialMessageable
    from ..threads import Thread
    from ..ext.application_commands import InvokableApplicationCommand

    InteractionChannel = Union[
        VoiceChannel, StageChannel, TextChannel, CategoryChannel, StoreChannel, Thread, PartialMessageable
    ]


class ApplicationCommandInteraction(Interaction):
    """Represents an interaction with an application command.

    Current examples are slash commands, user commands and message commands.

    .. versionadded:: 2.1

    Attributes
    -----------
    id: :class:`int`
        The interaction's ID.
    type: :class:`InteractionType`
        The interaction type.
    guild_id: Optional[:class:`int`]
        The guild ID the interaction was sent from.
    channel_id: Optional[:class:`int`]
        The channel ID the interaction was sent from.
    application_id: :class:`int`
        The application ID that the interaction was for.
    author: Optional[Union[:class:`User`, :class:`Member`]]
        The user or member that sent the interaction.
    token: :class:`str`
        The token to continue the interaction. These are valid
        for 15 minutes.
    data: :class:`ApplicationCommandInteractionData`
        The wrapped interaction data.
    """
    def __init__(self, *, data: InteractionPayload, state: ConnectionState):
        super().__init__(data=data, state=state)
        self.data = ApplicationCommandInteractionData(
            data=data.get('data'),
            state=state,
            guild=self.guild
        )
        self.application_command: InvokableApplicationCommand = None
        self.command_failed: bool = False

    @property
    def target(self) -> Optional[Union[User, Member, Message]]:
        return self.data.target
    
    @property
    def options(self) -> Optional[Dict[str, Any]]:
        return self.data.options


class ApplicationCommandInteractionData:
    """Represents the data of an interaction with an application command.

    .. versionadded:: 2.1

    Attributes
    ----------
    id: :class:`int`
        The application command ID.
    name: :class:`str`
        The application command name.
    type: :class:`ApplicationCommandType`
        The application command type.
    resolved: :class:`ApplicationCommandInteractionDataResolved`
        All resolved objects related to this interaction.
    options: Dict[:class:`str`, :class:`Any`]
        Slash command option names and values entered by a user.
    target_id: :class:`int`
        ID of the user or message targetted by a user or message command
    target: Union[:class:`User`, :class:`Member`, :class:`Message`]
        The user or message targetted by a user or message command
    """

    __slots__ = (
        'id',
        'name',
        'type',
        'target_id',
        'target',
        'resolved',
        'options'
    )

    def __init__(self, *, data: ApplicationCommandInteractionDataPayload, state: ConnectionState, guild: Guild):
        data = {} if data is None else data
        self.id: int = int(data['id'])
        self.name: str = data['name']
        self.type: ApplicationCommandType = try_enum(ApplicationCommandType, data['type'])
        self.resolved = ApplicationCommandInteractionDataResolved(
            data=data.get('resolved'),
            state=state,
            guild=guild
        )
        target_id = data.get('target_id')
        self.target_id: Optional[int] = None if target_id is None else int(target_id)
        self.target: Optional[Union[User, Member, Message]] = self.resolved.get(self.target_id)
        def option_payload_to_value(payload):
            value = payload.get('value')
            if value is not None:
                return self.resolved.get_with_type(value, payload['type'], value)
            return {
                option['name']: option_payload_to_value(option)
                for option in payload.get('options', [])
            }
        self.options: Dict[str, Any] = {
            option['name']: option_payload_to_value(option)
            for option in data.get('options', [])
        }


class ApplicationCommandInteractionDataResolved:
    """Represents the resolved data related to
    an interaction with an application command.

    .. versionadded:: 2.1

    Attributes
    ----------
    members: :class:`Dict[int, Member]`
        IDs and prtial members (missing 'deaf' and 'mute')
    users: :class:`Dict[int, User]`
        IDs and users
    roles: :class:`Dict[int, Role]`
        IDs and roles
    channels: :class:`Dict[int, Channel]`
        IDs and partial channels (only 'id', 'name' and 'permissions' are included)
    messages: :class:`Dict[int, Message]`
        IDs and messages
    """

    __slots__ = (
        'members',
        'users',
        'roles',
        'channels',
        'messages'
    )

    def __init__(self, *, data: ApplicationCommandInteractionDataResolvedPayload, state: ConnectionState, guild: Guild):
        data = data or {}

        self.members: Dict[int, Member] = {}
        self.users: Dict[int, User] = {}
        self.roles: Dict[int, Role] = {}
        self.channels: Dict[int, InteractionChannel] = {}
        self.messages: Dict[int, Message] = {}

        users = data.get('users', {})
        members = data.get('members', {})
        roles = data.get('roles', {})
        channels = data.get('channels', {})
        messages = data.get('messages', {})

        for str_id, user in users.items():
            user_id = int(str_id)
            member = members.get(str_id)
            if member is not None:
                self.members[user_id] = Member(
                    data={**member, 'user': user},
                    guild=guild,
                    state=state
                )
            else:
                self.users[user_id] = User(state=state, data=user)
        
        for str_id, role in roles.items():
            self.roles[int(str_id)] = Role(guild=guild, state=state, data=role)
        
        for str_id, channel in channels.items():
            factory, ch_type = _threaded_channel_factory(channel['type'])
            if factory:
                channel['position'] = 0
                self.channels[int(str_id)] = factory(guild=guild, state=state, data=channel)
        
        for str_id, message in messages.items():
            channel_id = int(message['channel_id'])
            channel = guild.get_channel(channel_id) if guild else None
            if channel is None:
                channel = state.get_channel(channel_id)
            self.messages[int(str_id)] = Message(state=state, channel=channel, data=message)
    
    def get_with_type(self, key: Any, option_type: OptionType, default: Any = None):
        if isinstance(option_type, int):
            option_type = try_enum(OptionType, option_type)
        if option_type is OptionType.mentionable:
            key = int(key)
            result = self.members.get(key)
            if result is not None:
                return result
            result = self.users.get(key)
            if result is not None:
                return result
            return self.roles.get(key, default)
        
        if option_type is OptionType.user:
            key = int(key)
            member = self.members.get(key)
            if member is not None:
                return member
            return self.users.get(key, default)

        if option_type is OptionType.channel:
            return self.channels.get(int(key), default)
        
        if option_type is OptionType.role:
            return self.roles.get(int(key), default)
        
        return default

    def get(self, key: int):
        if key is None:
            return None
        res = self.members.get(key)
        if res is not None:
            return res
        res = self.users.get(key)
        if res is not None:
            return res
        res = self.roles.get(key)
        if res is not None:
            return res
        res = self.channels.get(key)
        if res is not None:
            return res
        return self.messages.get(key)



from .enums import NSFWLevel, try_enum, VerificationLevel

# TODO, gotta add others features from https://mystb.in/AdvertisersExperiencesMothers.json


class Party:
    """A meta class which takes in the data in the form of a dict sent by the API, as the only argument, and 
    generates an invite link and other meta information from that.

    .. container:: operations
        
        .. describe:: str(x)

            Returns the complete invite link.
        
        .. describe:: x == y

            Checks if two invite links are equal.
    
    Attributes
    ----------
    code: :class:`str`
        The raw code without the ``discord.gg/`` prefix.
    max_uses: :class:`int`
        The maximum number of times the code can be used. ``0`` means unlimited.
    uses: :class:`int`
        The number of times the code has been used.
    max_age: :class:`int`
        The maximum age of the invite in seconds.
    temporary: :class:`bool`
        If the invite is temporary or permanent.
    guild_id: :class:`int`
        The ID of the guild, for whose voice channel, the invite was created.
    guild_name: :class:`int`
        The name of the guild, for whose voice channel, the invite was created.
    guild_description: :class:`str`
        The description of the guild, for whose voice channel, the invite was created.
    guild_splash: :class:`str`
        The splash of the guild, for whose voice channel, the invite was created.
    guild_banner: :class:`str`
        The banner of the guild, for whose voice channel, the invite was created.
    guild_icon: :class:`str`
        The icon of the guild, for whose voice channel, the invite was created.
    guild_verification_level: :class:`VerificationLevel`
        The verification level of the guild, for whose voice channel, the invite was created.
    guild_vanity_url: :class:`str`
        The vanity url of the guild, for whose voice channel, the invite was created. (If applicable)
    guild_is_nsfw: :class:`bool`
        The NSFW status of the guild, for whose voice channel, the invite was created.
    guild_nsfw_level: :class:`NSFWLevel`
        The NSFW level of the guild, for whose voice channel, the invite was created.
    

    """

    def __init__(self, data: dict):
        self.code: str = data['code']
        self.uses: int = data.get('uses')
        self.max_uses: int = data.get('max_uses')
        self.max_age: int = data.get('max_age')
        self.temporary: bool = data.get('temporary')

        # JSONS
        self.guild_info: dict = data.get("guild")
        self.channel_info: dict = data.get("channel")

    def __eq__(self, o: object) -> bool:
        # I am not really sure if it works, will want some reviews. I also don't know if this is even required...
        return isinstance(o, Party) and self.code == o.code

    @property
    def guild_id(self):
        return self.guild_info["id"]

    @property
    def guild_name(self):
        return self.guild_info["name"]

    @property
    def guild_description(self):
        return self.guild_info["description"]

    @property
    def guild_splash(self):
        return self.guild_info["splash"]

    @property
    def guild_banner(self):
        return self.guild_info["banner"]

    @property
    def guild_icon(self):
        return self.guild_info["icon"]

    @property
    def guild_verification_level(self):
        return try_enum(VerificationLevel, self.guild_info["verification_level"])

    @property
    def guild_vanity_url(self):
        return f'https://discord.gg/{self.guild_info["vanity_url_code"]}' or None

    @property
    def guild_is_nsfw(self):
        return self.guild_info["nsfw"]

    @property
    def guild_nsfw_level(self):
        return try_enum(NSFWLevel, self.guild_info["nsfw_level"])

    def __str__(self):
        return f"https://discord.gg/{self.code}"



import re
from typing import Dict, List, Optional, Union
from .enums import ApplicationCommandType, OptionType, try_enum, enum_if_int, try_enum_to_int
from .errors import InvalidArgument
from .role import Role
from .user import User


__all__ = (
    "application_command_factory",
    "ApplicationCommand",
    "SlashCommand",
    "UserCommand",
    "MessageCommand",
    "OptionChoice",
    "Option",
    "ApplicationCommandPermissions",
    "RawApplicationCommandPermission",
)


def application_command_factory(data: dict):
    cmd_type = try_enum(ApplicationCommandType, data.get("type", 1))
    if cmd_type is ApplicationCommandType.chat_input:
        return SlashCommand.from_dict(data)
    if cmd_type is ApplicationCommandType.user:
        return UserCommand.from_dict(data)
    if cmd_type is ApplicationCommandType.message:
        return MessageCommand.from_dict(data)


class OptionChoice:
    """
    Represents an option choice.

    Parameters
    ----------
    name: :class:`str`
        the name of the option choice (visible to users)
    value: Union[:class:`str`, :class:`int`]
        the value of the option choice
    """

    def __init__(self, name: str, value: Union[str, int]):
        self.name: str = name
        self.value: Union[str, int] = value

    def __repr__(self):
        return "<OptionChoice name='{0.name}' value={0.value}>".format(self)

    def __eq__(self, other):
        return (
            self.name == other.name and
            self.value == other.value
        )
    
    def to_dict(self) -> dict:
        return {
            'name': self.name,
            'value': self.value
        }


class Option:
    """
    Represents a slash command option.

    Parameters
    ----------
    name: :class:`str`
        option's name
    description: :class:`str`
        option's description
    type: :class:`OptionType`
        the option type, e.g. ``OptionType.USER``
    required: :class:`bool`
        whether this option is required or not
    choices: List[:class:`OptionChoice`]
        the list of option choices
    options: List[:class:`Option`]
        the list of sub options. You can only specify this parameter if
        the ``type`` is :class:`OptionType.SUB_COMMAND` or :class:`OptionType.SUB_COMMAND_GROUP`
    """

    __slots__ = ("name", "description", "type", "required", "choices", "options", "_choice_connectors")

    def __init__(
        self,
        name: str,
        description: str = None,
        type: OptionType = None,
        required: bool = False,
        choices: List[OptionChoice] = None,
        options: list = None
    ):
        assert name.islower(), f"Option name {name!r} must be lowercase"
        self.name: str = name
        self.description: str = description
        self.type: OptionType = enum_if_int(OptionType, type) or OptionType.string
        self.required: bool = required
        self.choices: List[OptionChoice] = choices or []
        self.options: List[Option] = options or []
        # self._choice_connectors = {}
        
        # for i, choice in enumerate(self.choices):
        #     if self.type == Type.INTEGER:
        #         if not isinstance(choice.value, int):
        #             self._choice_connectors[i] = choice.value
        #             choice.value = i
        #     elif self.type == Type.STRING:
        #         if not isinstance(choice.value, str):
        #             valid_value = f"option_choice_{i}"
        #             self._choice_connectors[valid_value] = choice.value
        #             choice.value = valid_value

    def __repr__(self):
        return (
            f'<Option name={self.name!r} description={self.description!r} '
            f'type={self.type!r} required={self.required!r} choices={self.choices!r} '
            f'options={self.options!r}>'
        )

    def __eq__(self, other):
        return (
            self.name == other.name and
            self.description == other.description and
            self.type == other.type and
            self.required == other.required and
            self.choices == other.choices and
            self.options == other.options
        )

    @classmethod
    def from_dict(cls, payload: dict):
        if 'options' in payload:
            payload['options'] = [Option.from_dict(p) for p in payload['options']]
        if 'choices' in payload:
            payload['choices'] = [OptionChoice(**p) for p in payload['choices']]
        return Option(**payload)

    def add_choice(self, name: str, value: Union[str, int]) -> None:
        """
        Adds an OptionChoice to the list of current choices
        Parameters are the same as for :class:`OptionChoice`
        """
        # Wrap the value
        # true_value = value
        # if self.type == OptionType.string:
        #     if not isinstance(value, str):
        #         true_value = f"option_choice_{len(self._choice_connectors)}"
        #         self._choice_connectors[true_value] = value
        # elif self.type == OptionType.integer:
        #     if not isinstance(value, int):
        #         true_value = len(self._choice_connectors)
        #         self._choice_connectors[true_value] = value
        # Add an option choice
        self.choices.append(OptionChoice(name=name, value=value))

    def add_option(
        self,
        name: str,
        description: str = None,
        type: OptionType = None,
        required: bool = False,
        choices: List[OptionChoice] = None,
        options: list = None
    ) -> None:
        """
        Adds an option to the current list of options
        Parameters are the same as for :class:`Option`
        """
        if self.type == 1:
            if type < 3:
                raise ValueError('sub_command can only be nested in a sub_command_group')
        elif self.type == 2:
            if type != 1:
                raise ValueError('Expected sub_command in this sub_command_group')
        self.options.append(
            Option(
                name=name,
                description=description,
                type=type,
                required=required,
                choices=choices,
                options=options
            )
        )

    def to_dict(self) -> dict:
        payload = {
            'name': self.name,
            'description': self.description,
            'type': try_enum_to_int(self.type)
        }
        if self.required:
            payload['required'] = True
        if len(self.choices) > 0:
            payload['choices'] = [c.to_dict() for c in self.choices]
        if len(self.options) > 0:
            payload['options'] = [o.to_dict() for o in self.options]
        return payload


class ApplicationCommand:
    """
    Base class for application commands
    """
    def __init__(self, type: ApplicationCommandType, **kwargs):
        self.type: ApplicationCommandType = enum_if_int(ApplicationCommandType, type)
        self.id: Optional[int] = kwargs.pop('id', None)
        if self.id:
            self.id = int(self.id)
        self.application_id: Optional[int] = kwargs.pop('application_id', None)
        if self.application_id:
            self.application_id = int(self.application_id)
        self._always_synced: bool = False
    
    def __eq__(self, other):
        return self._always_synced


class UserCommand(ApplicationCommand):
    def __init__(self, name: str, **kwargs):
        super().__init__(ApplicationCommandType.user, **kwargs)
        self.name: str = name
    
    def __repr__(self):
        return f"<UserCommand name={self.name!r}>"
    
    def __eq__(self, other):
        return self._always_synced or (
            self.type == other.type and
            self.name == other.name
        )

    def to_dict(self, **kwargs) -> dict:
        return {
            "type": try_enum_to_int(self.type),
            "name": self.name
        }
    
    @classmethod
    def from_dict(cls, data: dict):
        if data.pop("type", 0) == ApplicationCommandType.user.value:
            return UserCommand(**data)


class MessageCommand(ApplicationCommand):
    def __init__(self, name: str, **kwargs):
        super().__init__(ApplicationCommandType.message, **kwargs)
        self.name: str = name
    
    def __repr__(self):
        return f"<MessageCommand name={self.name!r}>"
    
    def __eq__(self, other):
        return self._always_synced or (
            self.type == other.type and
            self.name == other.name
        )

    def to_dict(self, **kwargs) -> dict:
        return {
            "type": try_enum_to_int(self.type),
            "name": self.name
        }
    
    @classmethod
    def from_dict(cls, data: dict):
        if data.pop("type", 0) == ApplicationCommandType.message.value:
            return MessageCommand(**data)


class SlashCommand(ApplicationCommand):
    """
    The base class for building slash commands.

    Parameters
    ----------
    name : :class:`str`
        The command name
    description : :class:`str`
        The command description (it'll be displayed by disnake)
    options : List[:class:`Option`]
        The options of the command. See :ref:`option`
    default_permission : :class:`bool`
        Whether the command is enabled by default when the app is added to a guild
    """

    def __init__(
        self,
        name: str,
        description: str,
        options: list = None,
        default_permission: bool = True,
        **kwargs
    ):
        super().__init__(ApplicationCommandType.chat_input, **kwargs)

        assert re.match(r"^[\w-]{1,32}$", name) is not None and name.islower(),\
            f"Slash command name {name!r} should consist of these symbols: a-z, 0-9, -, _"

        self.name: str = name
        self.description: str = description
        self.options: List[Option] = options or []
        self.default_permission: bool = default_permission
        self.permissions = ApplicationCommandPermissions()

    def __repr__(self):
        return (
            f'<SlashCommand name={self.name!r} description={self.description!r}'
            f'options={self.options!r} default_permission={self.default_permission!r}>'
        )

    def __eq__(self, other):
        return self._always_synced or (
            self.type == other.type and
            self.name == other.name and
            self.description == other.description and
            self.options == other.options
        )

    @classmethod
    def from_dict(cls, payload: dict):
        if payload.pop("type", 1) != ApplicationCommandType.chat_input.value:
            return None
        if 'options' in payload:
            payload['options'] = [Option.from_dict(p) for p in payload['options']]
        return SlashCommand(**payload)

    def add_option(
        self,
        name: str,
        description: str = None,
        type: int = None,
        required: bool = False,
        choices: List[OptionChoice] = None,
        options: list = None
    ) -> None:
        """
        Adds an option to the current list of options
        Parameters are the same as for :class:`Option`
        """
        self.options.append(
            Option(
                name=name,
                description=description,
                type=type,
                required=required,
                choices=choices,
                options=options
            )
        )

    def to_dict(self, *, hide_name=False) -> dict:
        res = {
            "type": try_enum_to_int(self.type),
            "description": self.description,
            "options": [o.to_dict() for o in self.options]
        }
        if not self.default_permission:
            res["default_permission"] = False
        if not hide_name:
            res["name"] = self.name
        return res


class RawApplicationCommandPermission:
    """
    Represents command permissions for a role or a user.
    Attributes
    ----------
    id : :class:`int`
        ID of a target
    type : :class:`int`
        1 if target is a role; 2 if target is a user
    permission : :class:`bool`
        allow or deny the access to the command
    """

    __slots__ = ("id", "type", "permission")

    def __init__(self, id: int, type: int, permission: bool):
        self.id: int = id
        self.type: int = type
        self.permission: bool = permission

    def __repr__(self):
        return "<RawCommandPermission id={0.id} type={0.type} permission={0.permission}>".format(self)

    @classmethod
    def from_pair(cls, target: Union[Role, User], permission: bool):
        if not isinstance(target, (Role, User)):
            raise InvalidArgument("target should be Role or User")
        if not isinstance(permission, bool):
            raise InvalidArgument("permission should be bool")
        return RawApplicationCommandPermission(
            id=target.id,
            type=1 if isinstance(target, Role) else 2,
            permission=permission
        )

    @classmethod
    def from_dict(cls, data: dict):
        return RawApplicationCommandPermission(
            id=data["id"],
            type=data["type"],
            permission=data["permission"]
        )

    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "type": self.type,
            "permission": self.permission
        }


class ApplicationCommandPermissions:
    """
    Represents application command permissions.
    Roughly equivalent to a list of :class:`RawApplicationCommandPermission`
    Application command permissions are checked on the server side.
    Only guild application commands can have this type of permissions.

    Parameters
    ----------
    raw_permissions : List[:class:`RawApplicationCommandPermission`]
        a list of :class:`RawApplicationCommandPermission`.
        However, :meth:`from_pairs` or :meth:`from_ids`
        might be more convenient.
    """

    def __init__(self, raw_permissions: List[RawApplicationCommandPermission] = None):
        self.permissions: List[RawApplicationCommandPermission] = raw_permissions or []

    def __repr__(self):
        return "<ApplicationCommandPermissions permissions={0.permissions!r}>".format(self)

    @classmethod
    def from_pairs(cls, permissions: Dict[Union[Role, User], bool]):
        """
        Creates :class:`ApplicationCommandPermissions` using
        instances of :class:`disnake.Role` and :class:`disnake.User`

        Parameters
        ----------
        permissions: Dict[Union[:class:`Role`, :class:`User`], :class:`bool`]
            a dictionary with permissions for users and roles.
        """
        raw_perms = [
            RawApplicationCommandPermission.from_pair(target, perm)
            for target, perm in permissions.items()
        ]

        return ApplicationCommandPermissions(raw_perms)

    @classmethod
    def from_ids(cls, role_perms: Dict[int, bool] = None, user_perms: Dict[int, bool] = None):
        """
        Creates :class:`ApplicationCommandPermissions` from
        2 dictionaries of IDs and permissions.

        Parameters
        ----------
        role_perms: Dict[:class:`int`, :class:`bool`]
            a dictionary of {``role_id``: :class:`bool`}
        user_perms: Dict[:class:`int`, :class:`bool`]
            a dictionary of {``user_id``: :class:`bool`}
        """
        role_perms = role_perms or {}
        user_perms = user_perms or {}
        raw_perms = [
            RawApplicationCommandPermission(role_id, 1, perm)
            for role_id, perm in role_perms.items()
        ]

        for user_id, perm in user_perms.items():
            raw_perms.append(RawApplicationCommandPermission(user_id, 2, perm))
        return ApplicationCommandPermissions(raw_perms)

    @classmethod
    def from_dict(cls, data: dict):
        return ApplicationCommandPermissions([
            RawApplicationCommandPermission.from_dict(perm)
            for perm in data["permissions"]
        ])

    def to_dict(self) -> dict:
        return {
            "permissions": [perm.to_dict() for perm in self.permissions]
        }



from __future__ import annotations
from typing import (
    List,
    Optional,
    TypeVar,
    TYPE_CHECKING
)
import asyncio
import datetime

from disnake.utils import async_all

from .cooldowns import BucketType, CooldownMapping, MaxConcurrency
from .errors import *


if TYPE_CHECKING:
    from typing_extensions import ParamSpec
    from disnake.interactions import ApplicationCommandInteraction
    from ._types import Check, Hook, Error
    from .cog import Cog


__all__ = ('InvokableApplicationCommand',)


T = TypeVar('T')
AppCommandT = TypeVar('AppCommandT', bound='InvokableApplicationCommand')
CogT = TypeVar('CogT', bound='Cog')
HookT = TypeVar('HookT', bound='Hook')
ErrorT = TypeVar('ErrorT', bound='Error')

if TYPE_CHECKING:
    P = ParamSpec('P')
else:
    P = TypeVar('P')


class InvokableApplicationCommand:
    """A base class that implements the protocol for a bot application command.

    These are not created manually, instead they are created via the
    decorator or functional interface.
    """

    def __init__(self, func, *, name: str = None, **kwargs):
        self.__command_flag__ = None
        self._callback = func
        self.name: str = name or func.__name__
        self.qualified_name: str = self.name
        if not isinstance(self.name, str):
            raise TypeError('Name of a command must be a string.')

        try:
            checks = func.__commands_checks__
            checks.reverse()
        except AttributeError:
            checks = kwargs.get('checks', [])

        self.checks: List[Check] = checks

        try:
            cooldown = func.__commands_cooldown__
        except AttributeError:
            cooldown = kwargs.get('cooldown')
        
        if cooldown is None:
            buckets = CooldownMapping(cooldown, BucketType.default)
        elif isinstance(cooldown, CooldownMapping):
            buckets = cooldown
        else:
            raise TypeError("Cooldown must be a an instance of CooldownMapping or None.")
        self._buckets: CooldownMapping = buckets

        try:
            max_concurrency = func.__commands_max_concurrency__
        except AttributeError:
            max_concurrency = kwargs.get('max_concurrency')
        self._max_concurrency: Optional[MaxConcurrency] = max_concurrency

        self.cog: Optional[CogT] = None
        self.guild_ids: List[int] = None
        self.auto_sync: bool = True

        self._before_invoke: Optional[HookT] = None
        self._after_invoke: Optional[HookT] = None
    
    @property
    def callback(self):
        return self._callback

    def add_check(self, func: Check) -> None:
        """Adds a check to the application command.

        This is the non-decorator interface to :func:`.check`.

        Parameters
        -----------
        func
            The function that will be used as a check.
        """

        self.checks.append(func)

    def remove_check(self, func: Check) -> None:
        """Removes a check from the application command.

        This function is idempotent and will not raise an exception
        if the function is not in the command's checks.

        Parameters
        -----------
        func
            The function to remove from the checks.
        """

        try:
            self.checks.remove(func)
        except ValueError:
            pass

    async def __call__(self, interaction: ApplicationCommandInteraction, *args, **kwargs) -> T:
        """|coro|

        Calls the internal callback that the application command holds.

        .. note::

            This bypasses all mechanisms -- including checks, converters,
            invoke hooks, cooldowns, etc. You must take care to pass
            the proper arguments and types to this function.

        """
        if self.cog is not None:
            return await self.callback(self.cog, interaction, *args, **kwargs)  # type: ignore
        else:
            return await self.callback(interaction, *args, **kwargs)  # type: ignore

    def _prepare_cooldowns(self, inter: ApplicationCommandInteraction) -> None:
        if self._buckets.valid:
            dt = inter.created_at
            current = dt.replace(tzinfo=datetime.timezone.utc).timestamp()
            bucket = self._buckets.get_bucket(inter, current)
            if bucket is not None:
                retry_after = bucket.update_rate_limit(current)
                if retry_after:
                    raise CommandOnCooldown(bucket, retry_after, self._buckets.type)  # type: ignore

    async def prepare(self, inter: ApplicationCommandInteraction) -> None:
        inter.application_command = self
        
        if not await self.can_run(inter):
            raise CheckFailure(f'The check functions for command {self.qualified_name!r} failed.')

        if self._max_concurrency is not None:
            await self._max_concurrency.acquire(inter)  # type: ignore

        try:
            self._prepare_cooldowns(inter)
        except:
            if self._max_concurrency is not None:
                await self._max_concurrency.release(inter)  # type: ignore
            raise
    
    def is_on_cooldown(self, inter: ApplicationCommandInteraction) -> bool:
        """Checks whether the application command is currently on cooldown.

        Parameters
        -----------
        inter: :class:`.ApplicationCommandInteraction`
            The interaction with the application command currently being invoked.

        Returns
        --------
        :class:`bool`
            A boolean indicating if the application command is on cooldown.
        """
        if not self._buckets.valid:
            return False

        bucket = self._buckets.get_bucket(inter)
        dt = inter.created_at
        current = dt.replace(tzinfo=datetime.timezone.utc).timestamp()
        return bucket.get_tokens(current) == 0

    def reset_cooldown(self, inter: ApplicationCommandInteraction) -> None:
        """Resets the cooldown on this application command.

        Parameters
        -----------
        inter: :class:`.ApplicationCommandInteraction`
            The interaction with this application command
        """
        if self._buckets.valid:
            bucket = self._buckets.get_bucket(inter)
            bucket.reset()

    def get_cooldown_retry_after(self, inter: ApplicationCommandInteraction) -> float:
        """Retrieves the amount of seconds before this application command can be tried again.

        Parameters
        -----------
        inter: :class:`.ApplicationCommandInteraction`
            The interaction with this application command.

        Returns
        --------
        :class:`float`
            The amount of time left on this command's cooldown in seconds.
            If this is ``0.0`` then the command isn't on cooldown.
        """
        if self._buckets.valid:
            bucket = self._buckets.get_bucket(inter)
            dt = inter.created_at
            current = dt.replace(tzinfo=datetime.timezone.utc).timestamp()
            return bucket.get_retry_after(current)

        return 0.0

    async def invoke(self, inter: ApplicationCommandInteraction, *args, **kwargs) -> None:
        """
        This method isn't really usable in this class, but it's usable in subclasses.
        """
        await self.prepare(inter)

        try:
            await self(inter, *args, **kwargs)
        except CommandError:
            inter.command_failed = True
            raise
        except asyncio.CancelledError:
            inter.command_failed = True
            return
        except Exception as exc:
            inter.command_failed = True
            raise CommandInvokeError(exc) from exc
        finally:
            if self._max_concurrency is not None:
                await self._max_concurrency.release(inter)

            await self.call_after_hooks(inter)

    def error(self, coro: ErrorT) -> ErrorT:
        """A decorator that registers a coroutine as a local error handler.

        A local error handler is an error event limited to a single application command.

        Parameters
        -----------
        coro: :ref:`coroutine <coroutine>`
            The coroutine to register as the local error handler.

        Raises
        -------
        TypeError
            The coroutine passed is not actually a coroutine.
        """

        if not asyncio.iscoroutinefunction(coro):
            raise TypeError('The error handler must be a coroutine.')

        self.on_error: Error = coro
        return coro

    def has_error_handler(self) -> bool:
        """
        Checks whether the application command has an error handler registered.
        """
        return hasattr(self, 'on_error')

    async def _call_local_error_handler(self, inter: ApplicationCommandInteraction, error: CommandError) -> None:
        if not self.has_error_handler():
            return
        if self.cog is None:
            await self.on_error(inter, error)
        else:
            await self.on_error(self.cog, inter, error)

    async def _call_external_error_handlers(self, inter: ApplicationCommandInteraction, error: CommandError) -> None:
        """Overwritten in subclasses"""
        raise error

    async def dispatch_error(self, inter: ApplicationCommandInteraction, error: CommandError):
        await self._call_local_error_handler(inter, error)
        await self._call_external_error_handlers(inter, error)

    async def call_before_hooks(self, inter: ApplicationCommandInteraction) -> None:
        # now that we're done preparing we can call the pre-command hooks
        # first, call the command local hook:
        cog = self.cog
        if self._before_invoke is not None:
            # should be cog if @commands.before_invoke is used
            instance = getattr(self._before_invoke, '__self__', cog)
            # __self__ only exists for methods, not functions
            # however, if @command.before_invoke is used, it will be a function
            if instance:
                await self._before_invoke(instance, inter)  # type: ignore
            else:
                await self._before_invoke(inter)  # type: ignore

        # TODO: call the cog local hook if applicable:
        # if cog is not None:
        #     hook = Cog._get_overridden_method(cog.cog_before_invoke)
        #     if hook is not None:
        #         await hook(ctx)

        # TODO: call the bot global hook if necessary
        # hook = ctx.bot._before_invoke
        # if hook is not None:
        #     await hook(ctx)

    async def call_after_hooks(self, inter: ApplicationCommandInteraction) -> None:
        cog = self.cog
        if self._after_invoke is not None:
            instance = getattr(self._after_invoke, '__self__', cog)
            if instance:
                await self._after_invoke(instance, inter)  # type: ignore
            else:
                await self._after_invoke(inter)  # type: ignore

        # TODO: call the cog local hook if applicable:
        # if cog is not None:
        #     hook = Cog._get_overridden_method(cog.cog_after_invoke)
        #     if hook is not None:
        #         await hook(ctx)

        # TODO: call the global hook
        # hook = ctx.bot._after_invoke
        # if hook is not None:
        #     await hook(ctx)

    def before_invoke(self, coro: HookT) -> HookT:
        """A decorator that registers a coroutine as a pre-invoke hook.

        A pre-invoke hook is called directly before the command is called.

        This pre-invoke hook takes a sole parameter, a :class:`.ApplicationCommandInteraction`.

        Parameters
        -----------
        coro: :ref:`coroutine <coroutine>`
            The coroutine to register as the pre-invoke hook.

        Raises
        -------
        TypeError
            The coroutine passed is not actually a coroutine.
        """
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError('The pre-invoke hook must be a coroutine.')

        self._before_invoke = coro
        return coro

    def after_invoke(self, coro: HookT) -> HookT:
        """A decorator that registers a coroutine as a post-invoke hook.

        A post-invoke hook is called directly after the command is called.

        This post-invoke hook takes a sole parameter, a :class:`.ApplicationCommandInteraction`.

        Parameters
        -----------
        coro: :ref:`coroutine <coroutine>`
            The coroutine to register as the post-invoke hook.

        Raises
        -------
        TypeError
            The coroutine passed is not actually a coroutine.
        """
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError('The post-invoke hook must be a coroutine.')

        self._after_invoke = coro
        return coro

    @property
    def cog_name(self) -> Optional[str]:
        """Optional[:class:`str`]: The name of the cog this application command belongs to, if any."""
        return type(self.cog).__cog_name__ if self.cog is not None else None

    async def can_run(self, inter: ApplicationCommandInteraction) -> bool:
        """|coro|

        Checks if the command can be executed by checking all the predicates
        inside the :attr:`~Command.checks` attribute.

        Parameters
        -----------
        inter: :class:`.ApplicationCommandInteraction`
            The interaction with the application command currently being invoked.

        Raises
        -------
        :class:`CommandError`
            Any application command error that was raised during a check call will be propagated
            by this function.

        Returns
        --------
        :class:`bool`
            A boolean indicating if the application command can be invoked.
        """

        original = inter.application_command
        inter.application_command = self

        try:
            # TODO: cog checks for application commands

            predicates = self.checks
            if not predicates:
                # since we have no checks, then we just return True.
                return True
            
            return await async_all(predicate(inter) for predicate in predicates)  # type: ignore
        finally:
            inter.application_command = original



from __future__ import annotations
from typing import List, TYPE_CHECKING, Callable

from .base_core import InvokableApplicationCommand
from .cog import Cog
from .errors import *

from disnake.app_commands import UserCommand, MessageCommand
from disnake._hub import _ApplicationCommandStore

import asyncio

if TYPE_CHECKING:
    from disnake.interactions import ApplicationCommandInteraction

__all__ = (
    'InvokableUserCommand',
    'InvokableMessageCommand',
    'user_command',
    'message_command'
)


class InvokableUserCommand(InvokableApplicationCommand):
    def __init__(self, func, *, name: str, guild_ids: List[int] = None, auto_sync: bool = True, **kwargs):
        super().__init__(func, name=name, **kwargs)
        self.guild_ids: List[int] = guild_ids
        self.auto_sync: bool = auto_sync
        self.body = UserCommand(name=self.name)
    
    async def _call_external_error_handlers(self, inter: ApplicationCommandInteraction, error: CommandError) -> None:
        cog = self.cog
        try:
            if cog is not None:
                local = Cog._get_overridden_method(cog.cog_user_command_error)
                if local is not None:
                    await local(inter, error)
        finally:
            inter.bot.dispatch('user_command_error', inter, error)


class InvokableMessageCommand(InvokableApplicationCommand):
    def __init__(self, func, *, name: str, guild_ids: List[int] = None, auto_sync: bool = True, **kwargs):
        super().__init__(func, name=name, **kwargs)
        self.guild_ids: List[int] = guild_ids
        self.auto_sync: bool = auto_sync
        self.body = MessageCommand(name=self.name)

    async def _call_external_error_handlers(self, inter: ApplicationCommandInteraction, error: CommandError) -> None:
        cog = self.cog
        try:
            if cog is not None:
                local = Cog._get_overridden_method(cog.cog_message_command_error)
                if local is not None:
                    await local(inter, error)
        finally:
            inter.bot.dispatch('message_command_error', inter, error)


def user_command(
    *,
    name: str = None,
    guild_ids: List[int] = None,
    auto_sync: bool = True,
    **kwargs
) -> Callable:
    """
    A decorator that builds a user command.

    Parameters
    ----------
    auto_sync: :class:`bool`
        whether to automatically register the command or not. Defaults to ``True``
    name: :class:`str`
        name of the user command you want to respond to (equals to function name by default).
    guild_ids: List[:class:`int`]
        if specified, the client will register the command in these guilds.
        Otherwise this command will be registered globally.
    """

    def decorator(func) -> InvokableUserCommand:
        if not asyncio.iscoroutinefunction(func):
            raise TypeError(f'<{func.__qualname__}> must be a coroutine function')
        new_func = InvokableUserCommand(
            func,
            name=name,
            guild_ids=guild_ids,
            auto_sync=auto_sync,
            **kwargs
        )
        _ApplicationCommandStore.user_commands[new_func.name] = new_func
        return new_func
    return decorator


def message_command(
    *,
    name: str = None,
    guild_ids: List[int] = None,
    auto_sync: bool = True,
    **kwargs
) -> Callable:
    """
    A decorator that builds a message command.

    Parameters
    ----------
    auto_sync: :class:`bool`
        whether to automatically register the command or not. Defaults to ``True``
    name: :class:`str`
        name of the message command you want to respond to (equals to function name by default).
    guild_ids: List[:class:`int`]
        if specified, the client will register the command in these guilds.
        Otherwise this command will be registered globally.
    """

    def decorator(func) -> InvokableMessageCommand:
        if not asyncio.iscoroutinefunction(func):
            raise TypeError(f'<{func.__qualname__}> must be a coroutine function')
        new_func = InvokableMessageCommand(
            func,
            name=name,
            guild_ids=guild_ids,
            auto_sync=auto_sync,
            **kwargs
        )
        _ApplicationCommandStore.message_commands[new_func.name] = new_func
        return new_func
    return decorator



from __future__ import annotations
from typing import TYPE_CHECKING, List, Dict, Tuple


if TYPE_CHECKING:
    from .ext.commands import (
        InvokableApplicationCommand,
        InvokableSlashCommand,
        InvokableUserCommand,
        InvokableMessageCommand
    )
    from disnake.app_commands import ApplicationCommand


def _get_all_commands() -> List[InvokableApplicationCommand]:
    all_commands = []
    for cmd in _ApplicationCommandStore.slash_commands.values():
        all_commands.append(cmd)
    for cmd in _ApplicationCommandStore.user_commands.values():
        all_commands.append(cmd)
    for cmd in _ApplicationCommandStore.message_commands.values():
        all_commands.append(cmd)
    return all_commands


def _ordered_unsynced_commands(
    test_guilds: List[int] = None
) -> Tuple[List[ApplicationCommand], Dict[int, List[ApplicationCommand]]]:
    global_cmds = []
    guilds = {}
    all_commands = _get_all_commands()
    for cmd in all_commands:
        if not cmd.auto_sync:
            cmd.body._always_synced = True
        guild_ids = cmd.guild_ids or test_guilds
        if guild_ids is None:
            global_cmds.append(cmd.body)
        else:
            for guild_id in guild_ids:
                if guild_id not in guilds:
                    guilds[guild_id] = [cmd.body]
                else:
                    guilds[guild_id].append(cmd.body)
    return global_cmds, guilds


class _ApplicationCommandStore:
    # I feel like this is a terrible solution,
    # but I don't know any exact reasons why.
    # If you know them, please tell me.
    slash_commands: Dict[str, InvokableSlashCommand] = {}
    user_commands: Dict[str, InvokableUserCommand] = {}
    message_commands: Dict[str, InvokableMessageCommand] = {}



from __future__ import annotations
from typing import Any, Dict, List, Tuple, Union, TYPE_CHECKING, Callable

from .base_core import InvokableApplicationCommand
from .cog import Cog
from .errors import *

from disnake.app_commands import SlashCommand, Option
from disnake.enums import OptionType
from disnake._hub import _ApplicationCommandStore

import asyncio

if TYPE_CHECKING:
    from disnake.interactions import ApplicationCommandInteraction

__all__ = (
    'InvokableSlashCommand',
    'SubCommandGroup',
    'SubCommand',
    'slash_command'
)


def options_as_route(options: Dict[str, Any]) -> Tuple[Tuple[str, ...], Dict[str, Any]]:
    if not options:
        return (), {}
    name, value = next(iter(options.items()))
    if isinstance(value, dict):
        chain, kwargs = options_as_route(value)
        return (name,) + chain, kwargs
    return (), options


class SubCommandGroup(InvokableApplicationCommand):
    def __init__(self, func, *, name: str = None, **kwargs):
        super().__init__(func, name=name, **kwargs)
        self.children: Dict[str, SubCommand] = {}
        self.option = Option(
            name=self.name,
            description='-',
            type=OptionType.sub_command_group,
            options=[]
        )
        self.qualified_name: str = None

    def sub_command(
        self,
        name: str = None,
        description: str = None,
        options: list = None,
        connectors: dict = None,
        **kwargs
    ) -> Callable:
        """
        A decorator that creates a subcommand in the
        subcommand group.
        Parameters are the same as in :class:`InvokableSlashCommand.sub_command`
        """

        def decorator(func) -> SubCommand:
            new_func = SubCommand(
                func,
                name=name,
                description=description,
                options=options,
                connectors=connectors,
                **kwargs
            )
            qualified_name = self.qualified_name or self.name
            new_func.qualified_name = f'{qualified_name} {new_func.name}'
            self.children[new_func.name] = new_func
            self.option.options.append(new_func.option)
            return new_func
        return decorator


class SubCommand(InvokableApplicationCommand):
    def __init__(
        self,
        func,
        *,
        name: str = None,
        description: str = None,
        options: list = None,
        connectors: Dict[str, str] = None,
        **kwargs
    ):
        super().__init__(func, name=name, **kwargs)
        self.connectors: Dict[str, str] = connectors
        self.option = Option(
            name=self.name,
            description=description or '-',
            type=OptionType.sub_command,
            options=options
        )
        self.qualified_name = None


class InvokableSlashCommand(InvokableApplicationCommand):
    def __init__(
        self,
        func,
        *,
        name: str = None,
        description: str = None,
        options: List[Option] = None,
        default_permission: bool = True,
        guild_ids: List[int] = None,
        connectors: Dict[str, str] = None,
        auto_sync: bool = True,
        **kwargs
    ):
        super().__init__(func, name=name, **kwargs)
        self.connectors: Dict[str, str] = connectors
        self.children: Dict[str, Union[SubCommand, SubCommandGroup]] = {}
        self.auto_sync: bool = auto_sync
        self.guild_ids: List[int] = guild_ids
        self.body = SlashCommand(
            name=self.name,
            description=description or '-',
            options=options or [],
            default_permission=default_permission,
        )
    
    def sub_command(
        self,
        name: str = None,
        description: str = None,
        options: list = None,
        connectors: dict = None,
        **kwargs
    ) -> Callable:
        """
        A decorator that creates a subcommand under the base command.

        Parameters
        ----------
        name: :class:`str`
            the name of the subcommand. Defaults to the function name
        description: :class:`str`
            the description of the subcommand
        options: List[:class:`Option`]
            the options of the subcommand for registration in API
        connectors: Dict[:class:`str`, :class:`str`]
            which function param states for each option. If the name
            of an option already matches the corresponding function param,
            you don't have to specify the connectors. Connectors template:
            ``{"option-name": "param_name", ...}``
        """
        def decorator(func) -> SubCommand:
            if len(self.children) == 0:
                if len(self.body.options) > 0:
                    self.body.options = []
            new_func = SubCommand(
                func,
                name=name,
                description=description,
                options=options,
                connectors=connectors,
                **kwargs
            )
            new_func.qualified_name = f'{self.qualified_name} {new_func.name}'
            self.children[new_func.name] = new_func
            self.body.options.append(new_func.option)
            return new_func
        return decorator

    def sub_command_group(
        self,
        name: str = None,
        **kwargs
    ) -> Callable:
        """
        A decorator that creates a subcommand group under the base command.
        Remember that the group must have at least one subcommand.

        Parameters
        ----------
        name : :class:`str`
            the name of the subcommand group. Defaults to the function name
        """
        def decorator(func) -> SubCommandGroup:
            if len(self.children) == 0:
                if len(self.body.options) > 0:
                    self.body.options = []
            new_func = SubCommandGroup(func, name=name, **kwargs)
            new_func.qualified_name = f'{self.qualified_name} {new_func.name}'
            self.children[new_func.name] = new_func
            self.body.options.append(new_func.option)
            return new_func
        return decorator

    async def _call_external_error_handlers(self, inter: ApplicationCommandInteraction, error: CommandError) -> None:
        cog = self.cog
        try:
            if cog is not None:
                local = Cog._get_overridden_method(cog.cog_slash_command_error)
                if local is not None:
                    await local(inter, error)
        finally:
            inter.bot.dispatch('slash_command_error', inter, error)

    async def invoke_children(self, inter: ApplicationCommandInteraction):
        chain, kwargs = options_as_route(inter.options)
        
        if len(chain) == 0:
            group = None
            subcmd = None
        elif len(chain) == 1:
            group = None
            subcmd = self.children.get(chain[0])
        elif len(chain) == 2:
            group = self.children.get(chain[0])
            subcmd = group.children.get(chain[1]) if group is not None else None

        if group is not None:
            try:
                await group.invoke(inter)
            except Exception as exc:
                await group._call_local_error_handler(inter, exc)
                raise
        
        if subcmd is not None:
            try:
                await subcmd.invoke(inter, **kwargs)
            except Exception as exc:
                await subcmd._call_local_error_handler(inter, exc)
                raise

    async def invoke(self, inter: ApplicationCommandInteraction):
        await self.prepare(inter)

        try:
            if len(self.children) > 0:
                await self(inter)
                await self.invoke_children(inter)
            else:
                await self(inter, **inter.options)
        except CommandError:
            inter.command_failed = True
            raise
        except asyncio.CancelledError:
            inter.command_failed = True
            return
        except Exception as exc:
            inter.command_failed = True
            raise CommandInvokeError(exc) from exc
        finally:
            if self._max_concurrency is not None:
                await self._max_concurrency.release(inter)

            await self.call_after_hooks(inter)


def slash_command(
    *,
    name: str = None,
    description: str = None,
    options: List[Option] = None,
    default_permission: bool = True,
    guild_ids: List[int] = None,
    connectors: Dict[str, str] = None,
    auto_sync: bool = True,
    **kwargs
) -> Callable:
    """
    A decorator that builds a slash command.

    Parameters
    ----------
    auto_sync: :class:`bool`
        whether to automatically register the command or not. Defaults to ``True``
    name: :class:`str`
        name of the slash command you want to respond to (equals to function name by default).
    description: :class:`str`
        the description of the slash command. It will be visible in Discord.
    options: List[:class:`Option`]
        the list of slash command options. The options will be visible in Discord.
    default_permission: :class:`bool`
        whether the command is enabled by default when the app is added to a guild.
    guild_ids: List[:class:`int`]
        if specified, the client will register a command in these guilds.
        Otherwise this command will be registered globally.
    connectors: Dict[:class:`str`, :class:`str`]
        binds function names to option names. If the name
        of an option already matches the corresponding function param,
        you don't have to specify the connectors. Connectors template:
        ``{"option-name": "param_name", ...}``
    """

    def decorator(func) -> InvokableSlashCommand:
        if not asyncio.iscoroutinefunction(func):
            raise TypeError(f'<{func.__qualname__}> must be a coroutine function')
        new_func = InvokableSlashCommand(
            func,
            name=name,
            description=description,
            options=options,
            default_permission=default_permission,
            guild_ids=guild_ids,
            connectors=connectors,
            auto_sync=auto_sync,
            **kwargs
        )
        _ApplicationCommandStore.slash_commands[new_func.name] = new_func
        return new_func
    return decorator
